from .version import VERSION_NUMBER

name = "pydotfiles"

__all__ = ['api', 'models', 'utils']

# Enables single-sourcing of the version number: https://packaging.python.org/guides/single-sourcing-package-version/
__version__ = VERSION_NUMBER
