from pathlib import Path

"""
Internal constants
"""
PYDOTFILES_CACHE_DIRECTORY = f"{str(Path.home())}/.pydotfiles"


"""
External defaults
"""
DEFAULT_PYDOTFILES_CONFIG_LOCAL_DIRECTORY = f"{str(Path.home())}/.dotfiles"
DEFAULT_CONFIG_REMOTE_REPO = "https://github.com/JasonYao/pydotfiles-basic.git"
